// (c) 1998 G. Nagler
#include <stdio.h>
#include <string.h>

static int option_i3again = 0;
static int option_quiet = 0;

static unsigned readword(FILE* f)
{
  unsigned v = 0;

  for (int i = 0; i< 2; i++)
  {
    int c = fgetc(f);

    v |= (unsigned(c) << (i*8));
  }
  return v;
}

static long readlong(FILE* f)
{
  long v = 0;

  for (int i = 0; i< 4; i++)
  {
    int c = fgetc(f);

    v |= (long(c) << (i*8));
  }
  return v;
}

int i5mtoi3(char* filename)
{
  int isbsq = (strstr(filename, ".bsq") != 0);
  FILE* f = fopen(filename, "rb+");
  if (!f)
  {
    if (!option_quiet)
      perror(filename);
    return 0;
  }
  char korgid[16];
  if (fread(korgid, 16, 1, f) != 1)
  {
    fprintf(stderr, "%s: read error\n", filename);
    fclose(f);
    return 0;
  }
  if (strncmp(korgid, "KORG9", 5) == 0 && isbsq && korgid[6] == 1 && korgid[7] == 1)
    if (!option_quiet) fprintf(stderr, "%s is already a Korg i3 file\n", filename);
  if (option_i3again == 0 && strncmp(korgid, "KORG9", 5) == 0 && !isbsq && korgid[7] == 0)
  {
    if (!option_quiet) fprintf(stderr, "%s is already a Korg i3 file\n", filename);
  }
  else if (strncmp(korgid, "KORG<", 5) == 0 ||
	   strncmp(korgid, "KORG9", 5) == 0 ||
	   strncmp(korgid, "KORG5", 5) == 0 ||
	   strncmp(korgid, "KORGH", 5) == 0)
  {
    fseek(f, 4, SEEK_SET);
    if (fputc('9', f) != '9')
      fprintf(stderr, "i5M->i3 conversion failed\n");

    if (isbsq)
    {
      fseek(f, 6, SEEK_SET);
      if (fputc(1, f) != 1)
	fprintf(stderr, "i5M->i3 conversion failed\n");
      fseek(f, 7, SEEK_SET);
      if (fputc(1, f) != 1)
	fprintf(stderr, "i5M->i3 conversion failed\n");
    }
    else
    {
      fseek(f, 7, SEEK_SET);
      if (fputc(0, f) != 0)
	fprintf(stderr, "i5M->i3 conversion failed\n");
    }

    if (korgid[4] == '<' || korgid[4] == '9')
    if (!isbsq)
    {
      fseek(f, 0x10, SEEK_SET);
      // check if program bank numbers are valid
      if (korgid[6] == 4)
      { // i3/i5m style

         readlong(f);
         readlong(f);
         long indexpos = readlong(f);
         long indexlen = readlong(f);
         fseek(f, indexpos, SEEK_SET);
         long stylepos[4], stylelen[4];
         stylepos[0] = indexpos + indexlen;
         for (int i = 0; i < 4; i++)
         {
           fseek(f, 4, SEEK_CUR);
           stylelen[i] = (long)readword(f);

           if (i+1 < 4)
             stylepos[i+1] = stylepos[i]+stylelen[i];
         }

         for (i = 0; i< 4;  i++)
         {
           fseek(f, stylepos[i]+0x26, SEEK_SET); // track parameters

           for (int j = 0; j < 6; j++)
           {
             int prog = fgetc(f);
             long pos = ftell(f);
             int bank = fgetc(f);
             int newbank = bank, newprog = prog;

             if (j <= 1)
             { // drum, perc
               newbank = 2;
               if (prog >= 8)
                 newprog = (j == 0 ? 0 : 5); // gm or perc
             }
             else
             {
               if (bank >= 2)
                 newbank = 0;
               else
                 newbank = bank;
             }
             if (prog != newprog)
             {
               printf("%s: incompatible program used, changed to GM prog\n", filename);
               fseek(f, pos-1, SEEK_SET);
               fputc(newprog, f);
               fseek(f, pos+1, SEEK_SET);
             }
             if (bank != newbank)
             {
               printf("%s: incompatible program bank used, changed to GM bank\n", filename);
               fseek(f, pos, SEEK_SET);
               fputc(newbank, f);
               fseek(f, pos+1, SEEK_SET);
             }

             fgetc(f); // vol
             fgetc(f); // balance
           }
         }
      }
      else if (korgid[6] == 3)
      { // arr
        readlong(f); // name index
        readlong(f);
        long arrpos = readlong(f);
        int arrcount =  (int)(readlong(f) / 131);

        for (int i = 0; i< arrcount; i++, arrpos += 131)
        {
          fseek(f, arrpos + 0x22, SEEK_SET);
          for (int j = 0; j < 8; j++)
          {
            fgetc(f); // prog
            long pos = ftell(f);
            int bank = fgetc(f);

            int newbank;

            if (j <= 1)
            { // drum, perc
              newbank = 2;
            }
            else
            {
              if (bank >= 2)
                newbank = 0;
              else
                newbank = bank;
            }
            if (bank != newbank)
            {
              printf("%s: incompatible program bank used, changed to GM bank\n", filename);
              fseek(f, pos, SEEK_SET);
              fputc(newbank, f);
              fseek(f, pos+1, SEEK_SET);
            }
            fseek(f, 6, SEEK_CUR);
          }
        }
      }

    }
  }
  else
  {
     fprintf(stderr, "%s: not a (valid) Korg i5M file\n", filename);
     fclose(f);
     return 0;
  }
  fclose(f);
  return 1;
}

#ifdef NOMAIN
int main_i5mtoi3(int argc, char** argv)
#else
int main(int argc, char** argv)
#endif
{
  argc--; argv++;

  while (argc > 0 && **argv == '-')
  {
    if (strcmp(*argv, "-again") == 0)
    {
      option_i3again = 1;
      argc--; argv++; continue;
    }
    if (strncmp(*argv, "-quiet", 2) == 0)
    {
      option_quiet = 1;
      argc--; argv++; continue;
    }
    fprintf(stderr, "invalid option %s\n", *argv);
    argc--; argv++;
  }
  if (argc == 0)
  {
    printf("usage: i5mtoi3 [-again] filename.ext\n");
    printf("-again\tconvert even if it already has i3 signature\n");
    printf("ext can be * (all files), pcg, sty, arr, sng, bsq\n");
    return 1;
  }
  if (strchr(*argv, '.') == 0 || strstr(*argv, ".*"))
  {
     char filename[80];
     strcpy(filename, *argv);

     char* p = strrchr(filename, '.');
     if (p)
        *p = 0;
     p = filename + strlen(filename);
     strcpy(p, ".sty");
     i5mtoi3(filename);
     strcpy(p, ".arr");
     i5mtoi3(filename);
     strcpy(p, ".pcg");
     i5mtoi3(filename);
     strcpy(p, ".bsq");
     i5mtoi3(filename);
     strcpy(p, ".sng");
     i5mtoi3(filename);
     return 0;
  }
  else
    return i5mtoi3(*argv);
}